"""Constants and enums"""
import typing
from enum import Enum

STRESS: typing.Set[str] = {"ˈ", "ˌ"}
"""Default stress characters"""

PUNCTUATION_MAP: typing.Mapping[str, str] = {";": ",", ":": ",", "?": ".", "!": "."}
"""Default punctuation simplification into short (,) and long (.) pauses"""


class BlankBetween(str, Enum):
    """Placement of blank tokens"""

    TOKENS = "tokens"
    """Blank between every token/phoneme"""

    WORDS = "words"
    """Blank between every word"""



#!/usr/bin/env python3
"""Tests for learn_phoneme_ids"""
import unittest
from collections import Counter

from phonemes2ids import learn_phoneme_ids


class LearnPhonemeIdsTestCase(unittest.TestCase):
    """Test cases for learn_phoneme_ids"""

    def test_abc(self):
        """Test basic learning"""
        word_phonemes = [["a"], ["b"], ["c"], ["b", "c", "a"]]
        all_phonemes = set()
        phoneme_counts = Counter()

        learn_phoneme_ids(
            word_phonemes=word_phonemes,
            all_phonemes=all_phonemes,
            all_phoneme_counts=phoneme_counts,
        )

        self.assertEqual(all_phonemes, {"a", "b", "c"})

        # Verify counts
        self.assertEqual(len(phoneme_counts), 3)

        for p in all_phonemes:
            self.assertEqual(phoneme_counts[p], 2)

    def test_simple_punctuation(self):
        """Test punctuation replacement with short/long pauses (,/.)"""
        word_phonemes = [[".", ",", ";", "!", "?"]]
        all_phonemes = set()

        # Without simple punctuation
        learn_phoneme_ids(word_phonemes=word_phonemes, all_phonemes=all_phonemes)
        self.assertEqual(all_phonemes, set(word_phonemes[0]))

        # With simple punctuation
        all_phonemes = set()
        learn_phoneme_ids(
            word_phonemes=word_phonemes,
            all_phonemes=all_phonemes,
            simple_punctuation=True,
        )
        self.assertEqual(all_phonemes, {",", "."})  # short/long pause

    def test_separate_graphemes(self):
        """Test grapheme separation (ɑ̃ -> \u0251, \u0303)"""
        word_phonemes = [["ɑ̃", "t͡ʃ"]]
        all_phonemes = set()

        # Without separate graphemes
        learn_phoneme_ids(word_phonemes=word_phonemes, all_phonemes=all_phonemes)
        self.assertEqual(all_phonemes, set(word_phonemes[0]))

        # With separate graphemes
        all_phonemes = set()
        learn_phoneme_ids(
            word_phonemes=word_phonemes,
            all_phonemes=all_phonemes,
            separate_graphemes=True,
        )
        self.assertEqual(all_phonemes, {"ɑ", "\u0303", "t", "\u0361", "ʃ"})

    def test_separate_tones(self):
        """Test tone separation (a1 -> a, 1)"""
        word_phonemes = [["a1", "b234"]]
        all_phonemes = set()

        # Without separate tones
        learn_phoneme_ids(word_phonemes=word_phonemes, all_phonemes=all_phonemes)
        self.assertEqual(all_phonemes, set(word_phonemes[0]))

        # With separate tones
        all_phonemes = set()
        learn_phoneme_ids(
            word_phonemes=word_phonemes, all_phonemes=all_phonemes, separate_tones=True
        )
        self.assertEqual(all_phonemes, {"a", "b", "1", "234"})

    def test_separate_stress(self):
        """Test stress separation (ˈa -> ˈ, a)"""
        word_phonemes = [["ˈa", "ˌb"]]
        all_phonemes = set()

        # Without separate stress
        learn_phoneme_ids(word_phonemes=word_phonemes, all_phonemes=all_phonemes)
        self.assertEqual(all_phonemes, set(word_phonemes[0]))

        # With separate stress
        all_phonemes = set()
        learn_phoneme_ids(
            word_phonemes=word_phonemes, all_phonemes=all_phonemes, separate=["ˈ", "ˌ"]
        )
        self.assertEqual(all_phonemes, {"ˈ", "a", "ˌ", "b"})

    def test_phoneme_map(self):
        """Test phoneme mapping"""
        word_phonemes = [["a", "c"]]
        all_phonemes = set()

        # Without phoneme map
        learn_phoneme_ids(word_phonemes=word_phonemes, all_phonemes=all_phonemes)
        self.assertEqual(all_phonemes, set(word_phonemes[0]))

        # With phoneme map
        all_phonemes = set()
        learn_phoneme_ids(
            word_phonemes=word_phonemes,
            all_phonemes=all_phonemes,
            phoneme_map={"a": "b", "c": "b"},
        )
        self.assertEqual(all_phonemes, {"b"})


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Tests for phonemes2ids"""
import unittest
from collections import Counter

from phonemes2ids import phonemes2ids, BlankBetween


class Phoneme2IdsTestCase(unittest.TestCase):
    """Test cases for phonemes2ids"""

    def test_basic(self):
        """Test basic mapping"""
        word_phonemes = [["a"], ["b"], ["c"], ["b", "c", "a"]]
        phoneme_to_id = {"a": 1, "b": 2, "c": 3}

        ids = phonemes2ids(word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id)

        self.assertEqual(ids, [1, 2, 3, 2, 3, 1])

    def test_blank_between_words(self):
        """Test blank symbol between words"""
        word_phonemes = [["a"], ["b"], ["c"], ["b", "c", "a"]]
        blank = "#"
        phoneme_to_id = {"a": 1, "b": 2, "c": 3, blank: 4}

        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            blank=blank,
            blank_between=BlankBetween.WORDS,
        )

        # between words
        self.assertEqual(ids, [4, 1, 4, 2, 4, 3, 4, 2, 3, 1, 4])

        # No blanks at start/end
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            blank=blank,
            blank_between=BlankBetween.WORDS,
            blank_at_start=False,
            blank_at_end=False,
        )

        self.assertEqual(ids, [1, 4, 2, 4, 3, 4, 2, 3, 1])

    def test_blank_between_tokens(self):
        """Test blank symbol between tokens"""
        word_phonemes = [["a"], ["b"], ["c"], ["b", "c", "a"]]
        blank = "#"
        phoneme_to_id = {"a": 1, "b": 2, "c": 3, blank: 4}

        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            blank=blank,
            blank_between=BlankBetween.TOKENS,
        )

        # between every phoneme (token)
        self.assertEqual(ids, [4, 1, 4, 2, 4, 3, 4, 2, 4, 3, 4, 1, 4])

        # No blanks at start/end
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            blank=blank,
            blank_between=BlankBetween.TOKENS,
            blank_at_start=False,
            blank_at_end=False,
        )

        self.assertEqual(ids, [1, 4, 2, 4, 3, 4, 2, 4, 3, 4, 1])

    def test_bos_eos(self):
        """Test bos/eos symbols"""
        word_phonemes = [["a"], ["b"], ["c"], ["b", "c", "a"]]
        bos = "^"
        eos = "$"
        phoneme_to_id = {"a": 1, "b": 2, "c": 3, bos: 4, eos: 5}

        ids = phonemes2ids(
            word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id, bos=bos, eos=eos
        )

        self.assertEqual(ids, [4, 1, 2, 3, 2, 3, 1, 5])

    def test_simple_punctuation(self):
        """Test punctuation replacement with short/long pauses (,/.)"""
        word_phonemes = [[".", ",", ";", "!", "?"]]
        phoneme_to_id = {".": 1, ",": 2, ";": 3, "!": 4, "?": 5}

        # Without simple punctuation
        ids = phonemes2ids(word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id)
        self.assertEqual(ids, [1, 2, 3, 4, 5])

        # With simple punctuation
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            simple_punctuation=True,
        )

        # ; -> ,
        # !/? -> .
        self.assertEqual(ids, [1, 2, 2, 1, 1])

    def test_separate_graphemes(self):
        """Test grapheme separation (ɑ̃ -> \u0251, \u0303)"""
        word_phonemes = [["ɑ̃", "t͡ʃ"]]
        phoneme_to_id = {
            "ɑ̃": 1,
            "t͡ʃ": 2,
            "ɑ": 3,
            "\u0303": 4,
            "t": 5,
            "\u0361": 6,
            "ʃ": 7,
        }

        # Without separate graphemes
        ids = phonemes2ids(word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id)
        self.assertEqual(ids, [1, 2])

        # With separate graphemes
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            separate_graphemes=True,
        )
        self.assertEqual(ids, [3, 4, 5, 6, 7])

    def test_separate_tones(self):
        """Test tone separation (a1 -> a, 1)"""
        word_phonemes = [["a1", "b234"]]
        phoneme_to_id = {"a1": 1, "b234": 2, "a": 3, "1": 4, "b": 5, "234": 6}

        # Without separate tones
        ids = phonemes2ids(word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id)
        self.assertEqual(ids, [1, 2])

        # With separate tones
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            separate_tones=True,
        )
        self.assertEqual(ids, [3, 4, 5, 6])

        # With separate tones (before phonemes: a1 -> 1, a)
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            separate_tones=True,
            tone_before=True,
        )
        self.assertEqual(ids, [4, 3, 6, 5])

    def test_separate_stress(self):
        """Test stress separation (ˈa -> ˈ, a)"""
        word_phonemes = [["ˈa", "ˌb"]]
        phoneme_to_id = {"ˈa": 1, "ˌb": 2, "ˈ": 3, "a": 4, "ˌ": 5, "b": 6}

        # Without separate stress
        ids = phonemes2ids(word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id)
        self.assertEqual(ids, [1, 2])

        # With separate stress
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            separate=["ˈ", "ˌ"],
        )
        self.assertEqual(ids, [3, 4, 5, 6])

    def test_phoneme_map(self):
        """Test phoneme mapping"""
        word_phonemes = [["a", "c"]]
        phoneme_to_id = {"a": 1, "b": 2, "c": 3}

        # Without phoneme map
        ids = phonemes2ids(word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id)
        self.assertEqual(ids, [1, 3])

        # With phoneme map
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            phoneme_map={"a": "b", "c": "b"},
        )
        self.assertEqual(ids, [2, 2])

    def test_missing_func(self):
        """Test function for determining missing ids"""
        word_phonemes = [["a", "b", "c"]]
        phoneme_to_id = {"_": 0, "a": 1, "c": 2}

        def missing_func(p):
            # ids for missing phoneme
            return [0]

        # Without missing func
        ids = phonemes2ids(word_phonemes=word_phonemes, phoneme_to_id=phoneme_to_id)
        self.assertEqual(ids, [1, 2])

        # With missing func
        ids = phonemes2ids(
            word_phonemes=word_phonemes,
            phoneme_to_id=phoneme_to_id,
            missing_func=missing_func,
        )
        self.assertEqual(ids, [1, 0, 2])


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""Utility methods for phonemes2ids"""
import typing


def load_phoneme_ids(phonemes_file: typing.TextIO) -> typing.Dict[str, int]:
    """
    Load phoneme id mapping from a text file.
    Format is ID<space>PHONEME
    Comments start with #

    Args:
        phonemes_file: text file

    Returns:
        dict with phoneme -> id
    """
    phoneme_to_id = {}
    for line in phonemes_file:
        line = line.strip("\r\n")
        if (not line) or line.startswith("#") or (" " not in line):
            # Exclude blank lines, comments, or malformed lines
            continue

        phoneme_id, phoneme_str = line.split(" ", maxsplit=1)
        phoneme_to_id[phoneme_str] = int(phoneme_id)

    return phoneme_to_id


def load_phoneme_map(
    phoneme_map_file: typing.TextIO,
) -> typing.Dict[str, typing.List[str]]:
    """
    Load phoneme/phoneme mapping from a text file.
    Format is FROM_PHONEME<space>TO_PHONEME[<space>TO_PHONEME...]
    Comments start with #

    Args:
        phoneme_map_file: text file

    Returns:
        dict with from_phoneme -> [to_phoneme, to_phoneme, ...]
    """
    phoneme_map = {}
    for line in phoneme_map_file:
        line = line.strip("\r\n")
        if (not line) or line.startswith("#") or (" " not in line):
            # Exclude blank lines, comments, or malformed lines
            continue

        from_phoneme, to_phonemes_str = line.split(" ", maxsplit=1)
        if not to_phonemes_str.strip():
            # To whitespace
            phoneme_map[from_phoneme] = [" "]
        else:
            # To one or more non-whitespace phonemes
            phoneme_map[from_phoneme] = to_phonemes_str.split()

    return phoneme_map



"""Setup file for phonemes2ids"""
import os
from pathlib import Path

import setuptools

this_dir = Path(__file__).parent
module_dir = this_dir / "phonemes2ids"

# -----------------------------------------------------------------------------

# Load README in as long description
long_description: str = ""
readme_path = this_dir / "README.md"
if readme_path.is_file():
    long_description = readme_path.read_text()

version_path = module_dir / "VERSION"
with open(version_path, "r", encoding="utf-8") as version_file:
    version = version_file.read().strip()

# -----------------------------------------------------------------------------

setuptools.setup(
    name="phonemes2ids",
    version=version,
    description="Convert phonemes to integer ids",
    author="Michael Hansen",
    author_email="mike@rhasspy.org",
    url="https://github.com/rhasspy/phonemes2ids",
    packages=setuptools.find_packages(),
    package_data={"phonemes2ids": ["VERSION", "py.typed"]},
    entry_points={"console_scripts": ["phonemes2ids = phonemes2ids.__main__:main",]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
)



"""Command-line interface to phonemes2ids"""
import argparse
import csv
import logging
import os
import sys
import typing
from collections import Counter

from . import learn_phoneme_ids, phonemes2ids
from .const import PUNCTUATION_MAP, STRESS, BlankBetween
from .utils import load_phoneme_ids, load_phoneme_map

_LOGGER = logging.getLogger("phonemes2ids")

# -----------------------------------------------------------------------------


def main():
    parser = argparse.ArgumentParser(prog="phonemes2ids")
    parser.add_argument(
        "--write-phonemes", help="Path to write phoneme ids text file (ID PHONEME)"
    )
    parser.add_argument(
        "--read-phonemes", help="Read phoneme ids from a text file (ID PHONEME)"
    )
    parser.add_argument(
        "-p", "--phoneme-separator", help="Separator character between phonemes"
    )
    parser.add_argument(
        "-w", "--word-separator", default=" ", help="Separator character between words"
    )
    parser.add_argument(
        "--id-separator", default=" ", help="Separator string each phoneme id"
    )
    parser.add_argument("--pad", help="Phoneme for padding (phoneme 0)")
    parser.add_argument("--bos", help="Phoneme to put at beginning of sentence")
    parser.add_argument("--eos", help="Phoneme to put at end of sentence")
    parser.add_argument(
        "--blank", help="Phoneme to put between words or tokens (see --blank-between)"
    )
    parser.add_argument(
        "--blank-between",
        choices=[v.value for v in BlankBetween],
        default=BlankBetween.WORDS,
        help="Where to insert blank phoneme (default: words)",
    )
    parser.add_argument(
        "--no-blank-start",
        action="store_true",
        help="Don't insert a blank token before the first word/token",
    )
    parser.add_argument(
        "--no-blank-end",
        action="store_true",
        help="Don't insert a blank token after the last word/token",
    )
    parser.add_argument(
        "--simple-punctuation",
        action="store_true",
        help="Map all punctuation into ',' and '.'",
    )
    parser.add_argument(
        "--csv",
        action="store_true",
        help="Input and output is CSV. Phonemes ids are added as a final column",
    )
    parser.add_argument(
        "--csv-delimiter", default="|", help="Delimiter in CSV input and output"
    )
    parser.add_argument(
        "--output-separator",
        default="|",
        help="Separator string between input phonemes and phoneme ids",
    )
    parser.add_argument(
        "--print-input", action="store_true", help="Print input text before phoneme ids"
    )
    parser.add_argument(
        "--separate-stress",
        action="store_true",
        help="Pull primary/secondary stress out as separate phonemes",
    )
    parser.add_argument(
        "--stress",
        action="append",
        help="Add phoneme to consider as stress (overwrites default of ˈˌ)",
    )
    parser.add_argument(
        "--separate-graphemes",
        action="store_true",
        help="Break apart graphemes into individual codepoints for phonemes",
    )
    parser.add_argument(
        "--separate-tones",
        action="store_true",
        help="Break apart tones (digits at the end of a phoneme) into individual phonemes",
    )
    parser.add_argument(
        "--tone-before",
        action="store_true",
        help="Insert separated tones before their corresponding phoneme instead of after",
    )
    parser.add_argument(
        "--separate",
        action="append",
        help="Break apart provided grapheme into separate phoneme",
    )
    parser.add_argument(
        "--write-phoneme-counts", help="Path to write phoneme counts observed in input"
    )
    parser.add_argument(
        "-m",
        "--map",
        nargs=2,
        action="append",
        help="Map from observed phoneme to desired phonemes",
    )
    parser.add_argument(
        "--phoneme-map",
        help="Path to text file with FROM_PHONEME TO_PHONEME on each line",
    )
    parser.add_argument("--version", action="store_true", help="Print version and exit")
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to the console"
    )
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # -------------------------------------------------------------------------

    if args.version:
        # Print version and exit
        from . import __version__

        print(__version__)
        sys.exit(0)

    # Map from observed phoneme to desired phonemes(s)
    phoneme_map: typing.Dict[str, typing.List[str]] = {}
    if args.phoneme_map:
        with open(args.phoneme_map, "r", encoding="utf-8") as phoneme_map_file:
            phoneme_map = load_phoneme_map(phoneme_map_file)

    if args.map:
        # Extra mappings from command line
        for from_phoneme, to_phonemes in args.map:
            if not to_phonemes.strip():
                # Whitespace
                phoneme_map[from_phoneme] = [" "]
            else:
                # Not whitespace
                phoneme_map[from_phoneme] = to_phonemes.split()

    phoneme_to_id: typing.Dict[str, int] = {}

    if args.read_phonemes:
        # Load from phonemes file
        # Format is ID<space>PHONEME
        with open(args.read_phonemes, "r", encoding="utf-8") as phonemes_file:
            phoneme_to_id.update(load_phoneme_ids(phonemes_file))

    if args.pad and (args.pad not in phoneme_to_id):
        # Add pad symbol
        phoneme_to_id[args.pad] = len(phoneme_to_id)

    if args.bos and (args.bos not in phoneme_to_id):
        # Add BOS symbol
        phoneme_to_id[args.bos] = len(phoneme_to_id)

    if args.eos and (args.eos not in phoneme_to_id):
        # Add EOS symbol
        phoneme_to_id[args.eos] = len(phoneme_to_id)

    if args.blank and (args.blank not in phoneme_to_id):
        # Add blank symbol
        phoneme_to_id[args.blank] = len(phoneme_to_id)

    separate: typing.Set[str] = set()
    if args.separate:
        separate.update(args.separate)

    if args.separate_stress:
        # Add stress symbols
        for stress in sorted(STRESS):
            separate.add(stress)
            if stress not in phoneme_to_id:
                phoneme_to_id[stress] = len(phoneme_to_id)

    # -------------------------------------------------------------------------

    if os.isatty(sys.stdin.fileno()):
        print("Reading phonemes from stdin...", file=sys.stderr)

    # CSV input/output
    if args.csv:
        csv_writer = csv.writer(sys.stdout, delimiter=args.csv_delimiter)
        reader = csv.reader(sys.stdin, delimiter=args.csv_delimiter)
    else:
        csv_writer = None
        reader = sys.stdin

    # Read all input and get set of phonemes
    all_phonemes = set(phoneme_to_id.keys())
    all_phoneme_counts = Counter()

    if args.simple_punctuation:
        # Add , and .
        all_phonemes.update(sorted(PUNCTUATION_MAP.values()))

    lines = []

    for line in reader:
        if args.csv:
            phonemes_str = line[-1]
        else:
            phonemes_str = line.strip()
            if not phonemes_str:
                continue

        # Split into words
        if args.phoneme_separator:
            word_phonemes = [
                word.split(args.phoneme_separator)
                for word in phonemes_str.split(args.word_separator)
            ]
        else:
            word_phonemes = [
                list(word) for word in phonemes_str.split(args.word_separator)
            ]

        lines.append((line, word_phonemes))

        # Accumulate phoneme set and counts
        learn_phoneme_ids(
            word_phonemes,
            all_phonemes,
            all_phoneme_counts=all_phoneme_counts,
            simple_punctuation=args.simple_punctuation,
            separate=separate,
            separate_graphemes=args.separate_graphemes,
            separate_tones=args.separate_tones,
            phoneme_map=phoneme_map,
        )

    # Assign phonemes to ids in sorted order
    for phoneme in sorted(all_phonemes):
        if phoneme not in phoneme_to_id:
            phoneme_to_id[phoneme] = len(phoneme_to_id)

    # -------------------------------------------------------------------------

    for line, word_phonemes in lines:
        if args.csv:
            phonemes_str = line[-1]
        else:
            phonemes_str = line.strip()

        # Transform into phoneme ids
        word_phoneme_ids = phonemes2ids(
            word_phonemes,
            phoneme_to_id=phoneme_to_id,
            pad=args.pad,
            bos=args.bos,
            eos=args.eos,
            blank=args.blank,
            blank_between=args.blank_between,
            blank_at_start=(not args.no_blank_start),
            blank_at_end=(not args.no_blank_end),
            simple_punctuation=args.simple_punctuation,
            separate=separate,
            separate_graphemes=args.separate_graphemes,
            separate_tones=args.separate_tones,
            tone_before=args.tone_before,
            phoneme_map=phoneme_map,
        )

        phoneme_ids_str = args.id_separator.join(
            (str(p_id) for p_id in word_phoneme_ids)
        )

        if args.csv:
            # Add phoneme ids as last column
            assert csv_writer is not None
            csv_writer.writerow((*line, phoneme_ids_str))
        else:
            if args.print_input:
                # Print input phonemes as well as phoneme ids
                print(phonemes_str, phoneme_ids_str, sep=args.output_separator)
            else:
                # Just print phoneme ids
                print(phoneme_ids_str)

    # -------------------------------------------------------------------------

    if args.write_phonemes:
        # Write file with ID<space>PHONEME format
        if args.write_phonemes == "-":
            print("")
            write_phoneme_ids(phoneme_to_id)
        else:
            # Write to file
            with open(args.write_phonemes, "w", encoding="utf-8") as phonemes_file:
                write_phoneme_ids(phoneme_to_id, phonemes_file)

    if args.write_phoneme_counts:
        # Write file with PHONEME<space>COUNT format
        if args.write_phoneme_counts == "-":
            print("")
            write_phoneme_counts(all_phoneme_counts)
        else:
            with open(args.write_phoneme_counts, "w", encoding="utf-8") as counts_file:
                write_phoneme_counts(all_phoneme_counts, counts_file)


# -----------------------------------------------------------------------------


def write_phoneme_ids(
    phoneme_to_id: typing.Mapping[str, int],
    phonemes_file: typing.Optional[typing.TextIO] = None,
):
    for phoneme, phoneme_id in sorted(phoneme_to_id.items(), key=lambda kv: kv[1]):
        print(phoneme_id, phoneme, file=phonemes_file)


def write_phoneme_counts(
    phoneme_counts: typing.Counter[str],
    counts_file: typing.Optional[typing.TextIO] = None,
):
    for phoneme, phoneme_count in phoneme_counts.most_common():
        print(phoneme, phoneme_count, file=counts_file)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



"""Tools for mapping phonemes to integer ids"""
import functools
import itertools
import logging
import operator
import typing
import unicodedata
from pathlib import Path

from .const import PUNCTUATION_MAP, STRESS, BlankBetween
from .utils import load_phoneme_ids, load_phoneme_map

_LOGGER = logging.getLogger("phoneme_ids")
_DIR = Path(__file__).parent

ID_LIST = typing.List[int]
WORD_ID_LIST = typing.List[ID_LIST]

__version__ = (_DIR / "VERSION").read_text().strip()

# -----------------------------------------------------------------------------


def phonemes2ids(
    word_phonemes: typing.List[typing.List[str]],
    phoneme_to_id: typing.Mapping[str, int],
    pad: typing.Optional[str] = None,
    bos: typing.Optional[str] = None,
    eos: typing.Optional[str] = None,
    blank: typing.Optional[str] = None,
    blank_between: typing.Union[str, BlankBetween] = BlankBetween.WORDS,
    blank_at_start: bool = True,
    blank_at_end: bool = True,
    simple_punctuation: bool = False,
    punctuation_map: typing.Optional[typing.Mapping[str, str]] = None,
    separate: typing.Optional[typing.Collection[str]] = None,
    separate_graphemes: bool = False,
    separate_tones: bool = False,
    tone_before: bool = False,
    phoneme_map: typing.Optional[typing.Mapping[str, typing.Sequence[str]]] = None,
    missing_func: typing.Optional[
        typing.Callable[[str], typing.Optional[typing.List[int]]]
    ] = None,
) -> ID_LIST:
    """
    Convert word-separated phonemes into integer ids.

    Args:
        word_phonemes: list of word phonemes, each a list of strings
        phoneme_to_id: map from phoneme to integer id
        pad: phoneme for padding vectors (currently unused)
        bos: phoneme to put at beginning of id list
        eos: phoneme to put at end of id list
        blank: phoneme to add between words or tokens
        blank_between: controls where blank tokens are inserted (see const.BlankBetween)
        blank_at_start: True if blank should also be inserted before first word/token
        blank_at_end: True if blank should also be inserted after last word/token
        simple_punctuation: True if punctuation should be simplified according to punctuation_map (see const.PUNCTUATION_MAP)
        punctuation_map: map from phoneme to phoneme, used when simple_punctuation is True
        separate: collection of phonemes that should be separated out into distinct phonemes (see const.STRESS)
        separate_graphemes: True if graphemes should be decomposed into codepoints as distinct phonemes
        separate_tones: True if digits at the end of phonemes (tones) should be separated out into distinct phonemes
        tone_before: True if tones separated out are inserted before their corresponding phoneme instead of after
        phoneme_map: optional map from phoneme to phoneme sequence (used after simplification/separation)
        missing_func: function called when phoneme is missing from phoneme_to_id map (str -> [int])

    Returns:
        ids - flat list of integer ids
    """
    if phoneme_map is None:
        phoneme_map = {}

    assert phoneme_map is not None

    if punctuation_map is None:
        punctuation_map = PUNCTUATION_MAP

    assert punctuation_map is not None

    is_separate: typing.Optional[typing.Callable[[str], bool]] = None
    if separate:
        is_separate = functools.partial(operator.contains, separate)

    blank_id: typing.Optional[int] = None
    if blank:
        blank_id = phoneme_to_id[blank]

    # Transform into phoneme ids
    word_phoneme_ids: WORD_ID_LIST = []

    def maybe_extend_ids(
        phoneme: str,
        target: typing.Union[ID_LIST, WORD_ID_LIST],
        append_list: bool = True,
    ):
        if not phoneme:
            return

        maybe_id = phoneme_to_id.get(phoneme)
        if maybe_id is not None:
            if append_list:
                typing.cast(WORD_ID_LIST, target).append([maybe_id])
            else:
                typing.cast(ID_LIST, target).append(maybe_id)
            return

        if missing_func is not None:
            maybe_ids = missing_func(phoneme)
            if maybe_ids:
                if append_list:
                    typing.cast(WORD_ID_LIST, target).append(maybe_ids)
                else:
                    typing.cast(ID_LIST, target).extend(maybe_ids)

    # Add beginning-of-sentence symbol
    if bos:
        maybe_extend_ids(bos, word_phoneme_ids)

    if (blank_id is not None) and blank_at_start:
        # Blank token at start
        word_phoneme_ids.append([blank_id])

    last_word_idx = len(word_phonemes) - 1
    for word_idx, word in enumerate(word_phonemes):
        word_ids: typing.List[int] = []

        if separate_graphemes:
            word = list(
                itertools.chain.from_iterable(
                    unicodedata.normalize("NFD", p) for p in word
                )
            )

        for phoneme in word:
            tone = ""

            if separate_tones:
                # Separate tones (digits at the end of a phoneme)
                tone_chars = []

                # Strip digits off the back of the phoneme (reversed)
                while phoneme and phoneme[-1].isdigit():
                    tone_chars.append(phoneme[-1])
                    phoneme = phoneme[:-1]

                if tone_chars:
                    tone = "".join(reversed(tone_chars))

                if tone and tone_before:
                    # Insert tone before corresponding phoneme
                    maybe_extend_ids(tone, word_ids, append_list=False)

            if is_separate is None:
                # No more splitting
                sub_phonemes = [phoneme]
            else:
                # Separate out stress, etc.
                assert separate is not None
                sub_phonemes = []

                before_split = ""
                for codepoint in phoneme:
                    if codepoint in separate:
                        # Split here
                        if before_split:
                            sub_phonemes.append(before_split)
                            before_split = ""

                        sub_phonemes.append(codepoint)
                    else:
                        before_split += codepoint

                if before_split:
                    sub_phonemes.append(before_split)

            for sub_phoneme in sub_phonemes:
                if not sub_phoneme:
                    continue

                if simple_punctuation and punctuation_map:
                    sub_phoneme = punctuation_map.get(sub_phoneme, sub_phoneme)

                to_phonemes = phoneme_map.get(sub_phoneme)
                if to_phonemes:
                    # Mapped to one or more phonemes
                    for to_phoneme in to_phonemes:
                        maybe_extend_ids(to_phoneme, word_ids, append_list=False)
                else:
                    # No map
                    maybe_extend_ids(sub_phoneme, word_ids, append_list=False)

            if tone and (not tone_before):
                # Insert tone after corresponding phoneme
                maybe_extend_ids(tone, word_ids, append_list=False)

        if word_ids:
            if blank_id is None:
                # No blank phoneme
                word_phoneme_ids.append(word_ids)
            elif blank_between == BlankBetween.TOKENS:
                # Blank phoneme between each token
                num_blanks = len(word_ids)
                word_phoneme_ids.append(
                    list(
                        # [p, blank, p, blank, ...]
                        itertools.chain.from_iterable(
                            # ((p, blank), (p, blank), ...)
                            itertools.zip_longest(
                                word_ids, itertools.repeat(blank_id, num_blanks)
                            )
                        )
                    )
                )

                if (
                    (word_idx == last_word_idx)
                    and (not blank_at_end)
                    and word_phoneme_ids[-1]
                ):
                    # Drop last blank
                    word_phoneme_ids[-1].pop()

            elif blank_between == BlankBetween.WORDS:
                # Blank phoneme between each word (list of tokens)
                word_phoneme_ids.append(word_ids)

                if (word_idx != last_word_idx) or blank_at_end:
                    word_phoneme_ids.append([blank_id])
            else:
                raise ValueError(f"Unexpected value for blank_between: {blank_between}")

    # Add end-of-sentence symbol
    if eos:
        maybe_extend_ids(eos, word_phoneme_ids)

    return list(itertools.chain.from_iterable(word_phoneme_ids))


# -----------------------------------------------------------------------------


def learn_phoneme_ids(
    word_phonemes: typing.List[typing.List[str]],
    all_phonemes: typing.Set[str],
    all_phoneme_counts: typing.Optional[typing.Counter[str]] = None,
    simple_punctuation: bool = False,
    punctuation_map: typing.Optional[typing.Mapping[str, str]] = None,
    separate: typing.Optional[typing.Collection[str]] = None,
    separate_graphemes: bool = False,
    separate_tones: bool = False,
    phoneme_map: typing.Optional[typing.Mapping[str, str]] = None,
):
    """
    Discover phonemes from examples.

    Args:
        word_phonemes: list of word phonemes, each a list of strings
        all_phonemes: set of distinct phonemes found (modified by function)
        all_phoneme_counts: optional Counter with observed phoneme counts (modified by function)
        simple_punctuation: True if punctuation should be simplified according to punctuation_map (see const.PUNCTUATION_MAP)
        punctuation_map: map from phoneme to phoneme, used when simple_punctuation is True
        separate: collection of phonemes that should be separated out into distinct phonemes (see const.STRESS)
        separate_graphemes: True if graphemes should be decomposed into codepoints as distinct phonemes
        separate_tones: True if digits at the end of phonemes (tones) should be separated out into distinct phonemes
        phoneme_map: optional map from phoneme to phoneme sequence (used after simplification/separation)

    Returns:
        None - all_phonemes and all_phoneme_counts are modified
    """
    if phoneme_map is None:
        phoneme_map = {}

    if punctuation_map is None:
        punctuation_map = PUNCTUATION_MAP

    is_separate: typing.Optional[typing.Callable[[str], bool]] = None
    if separate:
        is_separate = functools.partial(operator.contains, separate)

    for word in word_phonemes:
        if separate_graphemes:
            word = list(
                itertools.chain.from_iterable(
                    unicodedata.normalize("NFD", p) for p in word
                )
            )

        for phoneme in word:
            if separate_tones:
                # Separate tones (digits at the end of a phoneme)
                tone_chars = []

                # Strip digits off the back of the phoneme (reversed)
                while phoneme and phoneme[-1].isdigit():
                    tone_chars.append(phoneme[-1])
                    phoneme = phoneme[:-1]

                if tone_chars:
                    tone = "".join(reversed(tone_chars))
                    all_phonemes.add(tone)
                    if all_phoneme_counts is not None:
                        all_phoneme_counts[tone] += 1

            if is_separate is None:
                # No more splitting
                sub_phonemes = [phoneme]
            else:
                # Separate out stress, etc.
                assert separate is not None
                sub_phonemes = []

                before_split = ""
                for codepoint in phoneme:
                    if is_separate(codepoint):
                        # Split here
                        if before_split:
                            sub_phonemes.append(before_split)
                            before_split = ""

                        sub_phonemes.append(codepoint)
                    else:
                        before_split += codepoint

                if before_split:
                    sub_phonemes.append(before_split)

            for sub_phoneme in sub_phonemes:
                if not sub_phoneme:
                    continue

                if simple_punctuation:
                    sub_phoneme = punctuation_map.get(sub_phoneme, sub_phoneme)

                to_phonemes = phoneme_map.get(sub_phoneme)
                if to_phonemes:
                    # Mapped to one or more phonemes
                    for to_phoneme in to_phonemes:
                        all_phonemes.add(to_phoneme)

                        if all_phoneme_counts is not None:
                            all_phoneme_counts[to_phoneme] += 1
                else:
                    # No map
                    all_phonemes.add(sub_phoneme)

                    if all_phoneme_counts is not None:
                        all_phoneme_counts[sub_phoneme] += 1
